/* Build executable statement trees
   Copyright (C) 2000, 2001, 2002 Free Software Foundation, Inc.
   Contributed by Andy Vaught

This file is part of G95.

G95 is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

G95 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with G95; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/* st.c-- Executable statements are strung together into a singly
 * linked list of code structures.  These structures are later
 * translated into GBE tree structures and from there to executable
 * code for a target.  */

#include "g95.h"
#include <string.h>

g95_code new_st;



/* g95_clear_new_st()-- Zeroes out the new_st structure. */

void g95_clear_new_st(void) {

  memset(&new_st, '\0', sizeof(new_st));
  new_st.type = EXEC_NOP;
  new_st.where = g95_current_locus;
}



/* g95_add_statement()-- Takes the current new_st code structure and
 * adds it to the current program unit.  As a side-effect, it zeroes
 * the new_st. */

g95_code *g95_add_statement(void) {
g95_code *p;

  p = g95_get_code(-1, NULL);
  *p = new_st;

  p->where = g95_current_locus;

  if (g95_state_stack->head == NULL)
    g95_state_stack->head = p;

  *(g95_state_stack->next) = p;

  while(p->next != NULL)
    p = p->next;

  g95_state_stack->tail = p;
  g95_state_stack->next = &p->next;

  g95_clear_new_st();

  return p;
}



/* g95_append_code()-- Given some part of a g95_code structure, append
 * a set of code to its tail, returning a pointer to the new tail. */

g95_code *g95_append_code(g95_code *tail, g95_code *new) {

  if (tail != NULL) {
    while(tail->next != NULL)
      tail = tail->next;

    tail->next = new;
  }

  while(new->next != NULL)
    new = new->next;

  return new;
}



/* g95_get_code()-- Get a g95_code structure */

g95_code *g95_get_code(int type, g95_locus *where) {
g95_code *c;

  c = g95_getmem(sizeof(g95_code));
  c->type = type;
  c->where = (where == NULL) ? g95_current_locus : *where;
  return c;
}



/* free_statement()-- Free a single code structure, but not the actual
 * structure itself. */

static void free_statement(g95_code *p) {

  if (p->expr)
    g95_free_expr(p->expr);

  if (p->expr2)
    g95_free_expr(p->expr2);

  switch(p->type) {
  case EXEC_NOP:       case EXEC_ASSIGN:         case EXEC_GOTO:
  case EXEC_CYCLE:     case EXEC_RETURN:         case EXEC_STOP:
  case EXEC_EXIT:      case EXEC_PAUSE:          case EXEC_WHERE:
  case EXEC_IOLENGTH:  case EXEC_CONTINUE:       case EXEC_ENTRY:
  case EXEC_DO_WHILE:  case EXEC_ARITHMETIC_IF:  case EXEC_POINTER_ASSIGN:
  case EXEC_TRANSFER:  case EXEC_AC_ASSIGN:      case EXEC_AC_START:
  case EXEC_LABEL_ASSIGN: case EXEC_WHERE_ASSIGN:
    break;

  case EXEC_CALL:
    g95_free_actual_arglist(p->ext.sub.actual);
    break;

  case EXEC_SELECT:
    if (p->ext.case_list) g95_free_case_list(p->ext.case_list);
    break;

  case EXEC_DO:
    g95_free_iterator(p->ext.iterator, 1);
    break;

  case EXEC_ALLOCATE:
  case EXEC_DEALLOCATE:
    g95_free_alloc_list(p->ext.alloc_list);
    break;

  case EXEC_OPEN:
    g95_free_open(p->ext.open);
    break;

  case EXEC_CLOSE:
    g95_free_close(p->ext.close);
    break;

  case EXEC_FLUSH:
    g95_free_flush(p->ext.flush);
    break;

  case EXEC_WAIT:
    g95_free_wait(p->ext.wait);
    break;

  case EXEC_BACKSPACE:
  case EXEC_ENDFILE:
  case EXEC_REWIND:
    g95_free_filepos(p->ext.filepos);
    break;

  case EXEC_IF:
    g95_free_statements(p->ext.block);
    break;

  case EXEC_INQUIRE:
    g95_free_inquire(p->ext.inquire);
    break;

  case EXEC_READ:
  case EXEC_WRITE:
    g95_free_dt(p->ext.dt);
    break;

  case EXEC_DT_END:
    /* The ext.dt member is a duplicate pointer and doesn't need to be freed */
    break;

  case EXEC_FORALL:
    g95_free_forall_iterator(p->ext.forall_iterator);
    break;

  default:
    g95_internal_error("free_statement(): Bad statement");
  }
}



/* g95_free_statements()-- Free a code statement and all other code
 * structures linked to it. */

void g95_free_statements(g95_code *p) {
g95_code *q;

  for(; p; p=q) {
    q = p->next;

    if (p->block) g95_free_statements(p->block);
    free_statement(p);
    g95_free(p);
  }
}



/* g95_undo_statement()-- Frees everything associated with the current
 * statement.  */

void g95_undo_statement(void) {

  g95_free_statements(new_st.block);
  g95_free_statements(new_st.next);

  free_statement(&new_st);
  g95_clear_new_st();
}



/* code_indent()-- Do indentation for a specific level */

static void code_indent(int level, g95_st_label *label) {
int i;

  if (label != NULL)
    g95_status("%-6d; ", label->value);
  else
    g95_status("        ");

  for(i=0; i<2*level; i++)
    g95_status_char(' ');
}



/* g95_show_code_node()-- Show a single code node and everything
 * underneath it if necessary. */

static void g95_show_code_node(int level, g95_code *c) {
g95_forall_iterator *fa;
char *module, *name;
g95_open *open;
g95_case *cp;
g95_alloc *a;
g95_code *d;
g95_close *close;
g95_filepos *fp;
g95_inquire *i;
g95_dt *dt;
int comma;

  code_indent(level, c->here); 

  switch(c->type) {
  case EXEC_NOP:
    g95_status("nop(");
    comma = 0;
    break;

  case EXEC_CONTINUE:
    g95_status("cont(");
    comma = 0;
    break;

  case EXEC_AC_START:
    g95_status("ac_start(sym='%s'", c->sym->name);
    comma = 1;
    break;

  case EXEC_AC_ASSIGN:
    g95_status("ac_assign(sym='%s', expr=", c->sym->name);
    g95_show_expr(c->expr);
    break;

  case EXEC_ASSIGN:
    g95_status("assign(lhs=");
    g95_show_expr(c->expr);
    g95_status(", rhs=");
    g95_show_expr(c->expr2);
    comma = 1;
    break;

  case EXEC_WHERE_ASSIGN:
    g95_status("where_assign(lhs=");
    g95_show_expr(c->expr);
    g95_status(", rhs=");
    g95_show_expr(c->expr2);
    comma = 1;
    break;

  case EXEC_POINTER_ASSIGN:
    g95_status("ptr_assign(lhs=");
    g95_show_expr(c->expr);
    g95_status(", rhs=");
    g95_show_expr(c->expr2);
    comma = 1;
    break;

  case EXEC_GOTO:
    if (c->label != NULL)
      g95_status("goto(label=%d", c->label->value);
    else
      g95_status("goto(var='%s'", c->sym->name);

    comma = 1;
    break;

  case EXEC_CALL:
    if (c->sym == NULL)
      module = name = "";
    else {
      module = c->sym->module;
      name   = c->sym->name;
    }

    g95_status("call(name='%s:%s', sub='%s', arg=",
	       module, name, c->ext.sub.sub_name);
    g95_show_actual_arglist(c->ext.sub.actual);
    comma = 1;
    break;

  case EXEC_RETURN:
    g95_status("ret(");
    comma = 0;

    if (c->expr) {
      g95_status("value=");
      g95_show_expr(c->expr);
      comma = 1;
    }

    break;

  case EXEC_STOP:
    g95_status("stop(");

    if (c->expr != NULL) {
      g95_status("expr=");
      g95_show_expr(c->expr);
    } else
      g95_status("code=%d", c->ext.stop_code);

    comma = 1;
    break;

  case EXEC_PAUSE:
    g95_status("pause(");

    if (c->expr != NULL) {
      g95_status("expr=");
      g95_show_expr(c->expr);
    } else
      g95_status("code=%d", c->ext.stop_code);

    comma = 1;
    break;

  case EXEC_ARITHMETIC_IF:
    g95_status("arith_if(expr=");

    g95_show_expr(c->expr);

    g95_status(", lt0=%d, eq0=%d, gt0=%d",
	       c->label->value, c->label2->value, c->label3->value);

    comma = 1;
    break;

  case EXEC_IF:
    g95_status("log_if(expr=");
    g95_show_expr(c->expr);

    g95_status(", true=\n");
    g95_show_code(level+1, c->block);

    if (c->ext.block != NULL) {
      code_indent(level, 0);
      g95_status(", else=\n");
      g95_show_code(level+1, c->ext.block);
    }

    code_indent(level, c->label);
    g95_status(") # ENDIF\n");
    break;

  case EXEC_SELECT:
    d = c->block;
    g95_status("select(expr=");
    g95_show_expr((c->expr != NULL) ? c->expr : c->expr2);
    g95_status(", cases=[\n");

    for(;d ;d=d->block) {
      code_indent(level, 0);

      g95_status("case(");
      for(cp=d->ext.case_list; cp; cp=cp->next) {
	g95_status_char('(');
	g95_show_expr(cp->low);
	g95_status_char(' ');
	g95_show_expr(cp->high);
	g95_status_char(')');
	g95_status_char(' ');
      }

      g95_status_char('\n');

      g95_show_code(level+1, d->next);
    }

    code_indent(level, c->label);
    g95_status("END SELECT");
    break;

  case EXEC_WHERE:
    g95_status("WHERE ");

    d = c->block;
    g95_show_expr(d->expr);
    g95_status_char('\n');

    g95_show_code(level+1, d->next);

    for(d=d->block; d; d=d->block) {
      code_indent(level, 0);
      g95_status("ELSE WHERE ");
      g95_show_expr(d->expr);
      g95_status_char('\n');
      g95_show_code(level+1, d->next);
    }

    code_indent(level, 0);
    g95_status("END WHERE");
    break;


  case EXEC_FORALL:
    g95_status("FORALL ");
    for(fa=c->ext.forall_iterator; fa; fa=fa->next) {
      g95_show_expr(fa->var);
      g95_status_char(' ');
      g95_show_expr(fa->start);
      g95_status_char(':');
      g95_show_expr(fa->end);
      g95_status_char(':');
      g95_show_expr(fa->stride);

      if (fa->next != NULL)
	g95_status_char(',');
    }

    if (c->expr != NULL) {
      g95_status_char(',');
      g95_show_expr(c->expr);
    }
    g95_status_char('\n');

    g95_show_code(level+1, c->block);

    code_indent(level, 0);
    g95_status("END FORALL");
    break;
    
  case EXEC_DO:
    g95_status("do(var=");
    g95_show_expr(c->ext.iterator->var);

    g95_status(", start=");
    g95_show_expr(c->ext.iterator->start);

    g95_status(", end=");
    g95_show_expr(c->ext.iterator->end);

    g95_status(", step=");
    g95_show_expr(c->ext.iterator->step);

    g95_status(", body=\n");
    g95_show_code(level+1, c->block);

    code_indent(level, 0);
    comma = 1;
    break;

  case EXEC_DO_WHILE:
    g95_status("do_while(expr=");
    g95_show_expr(c->expr);
    g95_status(", body=");

    g95_show_code(level+1, c->block);

    code_indent(level, c->label);
    break;

  case EXEC_CYCLE:
    g95_status("cycle(");
    comma = 0;

    if (c->sym) {
      g95_status("label='%s'", c->sym->name);
      comma = 1;
    }

    break;

  case EXEC_EXIT:
    g95_status("exit(");
    comma = 0;

    if (c->sym) {
      g95_status("label='%s'", c->sym->name);
      comma = 1;
    }

    break;

  case EXEC_ALLOCATE:
    g95_status("allocate(");
    comma = 0;

    if (c->expr) {
      g95_status("stat=");
      g95_show_expr(c->expr);
      comma = 1;
    }

    if (comma)
      g95_status(", ");

    g95_status("alloc=[");

    for(a=c->ext.alloc_list; a; a=a->next) {
      g95_show_expr(a->expr);
      if (a->next != NULL)
	g95_status(", ");
    }

    g95_status_char(']');
    comma = 1;
    break;

  case EXEC_DEALLOCATE:
    g95_status("deallocate(");
    comma = 0;

    if (c->expr) {
      g95_status("stat=");
      g95_show_expr(c->expr);
      comma = 1;
    }

    if (comma)
      g95_status(", ");

    g95_status("alloc=[");

    for(a=c->ext.alloc_list; a; a=a->next) {
      g95_show_expr(a->expr);
      if (a->next != NULL)
	g95_status(", ");
    }

    g95_status_char(']');
    comma = 1;
    break;

  case EXEC_OPEN:
    g95_status("open_(");
    open = c->ext.open;
    comma = 0;

    if (open->unit) {
      if (comma) g95_status(", ");
      g95_status("unit="); g95_show_expr(open->unit); comma = 1;
    }

    if (open->iostat) {
      if (comma) g95_status(", ");
      g95_status("iostat="); g95_show_expr(open->iostat); comma = 1;
    }

    if (open->file) {
      if (comma) g95_status(", ");
      g95_status("file="); g95_show_expr(open->file); comma = 1;
    }

    if (open->status) {
      if (comma) g95_status(", ");
      g95_status("status="); g95_show_expr(open->status); comma = 1;
    }

    if (open->access) {
      if (comma) g95_status(", ");
      g95_status("access="); g95_show_expr(open->access); comma = 1;
    }

    if (open->form) {
      if (comma) g95_status(", ");
      g95_status("form="); g95_show_expr(open->form); comma = 1;
    }

    if (open->recl) {
      if (comma) g95_status(", ");
      g95_status("recl="); g95_show_expr(open->recl); comma = 1;
    }

    if (open->blank) {
      if (comma) g95_status(", ");
      g95_status("blank="); g95_show_expr(open->blank); comma = 1;
    }

    if (open->position) {
      if (comma) g95_status(", ");
      g95_status("position="); g95_show_expr(open->position); comma = 1;
    }

    if (open->action) {
      if (comma) g95_status(", ");
      g95_status("action="); g95_show_expr(open->action); comma = 1;
    }

    if (open->delim) {
      if (comma) g95_status(", ");
      g95_status("delim="); g95_show_expr(open->delim); comma = 1;
    }

    if (open->pad) {
      if (comma) g95_status(", ");
      g95_status("pad="); g95_show_expr(open->pad); comma = 1;
    }

    if (open->err != NULL) {
      if (comma) g95_status(", ");
      g95_status("err=%d", open->err->value); comma = 1;
    }

    break;

  case EXEC_CLOSE:
    g95_status("close(");
    close = c->ext.close;
    comma = 0;

    if (close->unit) {
      if (comma) g95_status(", ");
      g95_status("unit="); g95_show_expr(close->unit); comma = 1;
    }

    if (close->iostat) {
      if (comma) g95_status(", ");
      g95_status("iostat="); g95_show_expr(close->iostat); comma = 1;
    }

    if (close->status) {
      if (comma) g95_status(", ");
      g95_status("status=");g95_show_expr(close->status);  comma = 1;
    }

    if (close->err != NULL) {
      if (comma) g95_status(", ");
      g95_status("err=%d", close->err->value); comma = 1;
    }

    break;

  case EXEC_BACKSPACE:
    g95_status("backspace(");
    goto show_filepos;

  case EXEC_ENDFILE:
    g95_status("endfile(");
    goto show_filepos;

  case EXEC_REWIND:
    g95_status("rewind(");

  show_filepos:
    fp = c->ext.filepos;
    comma = 0;

    if (fp->unit) {
      if (comma) g95_status(", ");
      g95_status("unit="); g95_show_expr(fp->unit); comma = 1;
    }

    if (fp->iostat) {
      if (comma) g95_status(", ");
      g95_status("iostat="); g95_show_expr(fp->iostat); comma = 1;
    }

    if (fp->err != NULL) {
      if (comma) g95_status(", ");
      g95_status("err=%d", fp->err->value); comma = 1;
    }

    break;

  case EXEC_INQUIRE:
    g95_status("inquire(");
    i = c->ext.inquire;
    comma = 0;

    if (i->unit) {
      if (comma) g95_status(", ");
      g95_status("unit="); g95_show_expr(i->unit); comma = 1;
    }

    if (i->file) {
      if (comma) g95_status(", ");
      g95_status("file="); g95_show_expr(i->file); comma = 1;
    }

    if (i->iostat) {
      if (comma) g95_status(", ");
      g95_status("iostat="); g95_show_expr(i->iostat); comma = 1;
    }

    if (i->exist) {
      if (comma) g95_status(", ");
      g95_status("exist="); g95_show_expr(i->exist); comma = 1;
    }

    if (i->opened) {
      if (comma) g95_status(", ");
      g95_status("opened="); g95_show_expr(i->opened); comma = 1;
    }

    if (i->number) {
      if (comma) g95_status(", ");
      g95_status("number="); g95_show_expr(i->number); comma = 1;
    }

    if (i->named) {
      if (comma) g95_status(", ");
      g95_status("named="); g95_show_expr(i->named); comma = 1;
    }

    if (i->name) {
      if (comma) g95_status(", ");
      g95_status("name="); g95_show_expr(i->name); comma = 1;
    }

    if (i->access) {
      if (comma) g95_status(", ");
      g95_status("access="); g95_show_expr(i->access); comma = 1;
    }

    if (i->sequential) {
      if (comma) g95_status(", ");
      g95_status("sequential="); g95_show_expr(i->sequential); comma = 1;
    }

    if (i->direct) {
      if (comma) g95_status(", ");
      g95_status("direct="); g95_show_expr(i->direct); comma = 1;
    }

    if (i->form) {
      if (comma) g95_status(", ");
      g95_status("form="); g95_show_expr(i->form); comma = 1;
    }

    if (i->formatted) {
      if (comma) g95_status(", ");
      g95_status("formatted="); g95_show_expr(i->formatted); comma = 1;
    }

    if (i->unformatted) {
      if (comma) g95_status(", ");
      g95_status("unformatted="); g95_show_expr(i->unformatted); comma = 1;
    }

    if (i->recl) {
      if (comma) g95_status(", ");
      g95_status("recl="); g95_show_expr(i->recl); comma = 1;
    }

    if (i->nextrec) {
      if (comma) g95_status(", ");
      g95_status("nextrec="); g95_show_expr(i->nextrec); comma = 1;
    }

    if (i->blank) {
      if (comma) g95_status(", ");
      g95_status("blank="); g95_show_expr(i->blank); comma = 1;
    }

    if (i->position) {
      if (comma) g95_status(", ");
      g95_status("position="); g95_show_expr(i->position); comma = 1;
    }

    if (i->action) {
      if (comma) g95_status(", ");
      g95_status("action="); g95_show_expr(i->action); comma = 1;
    }

    if (i->read) {
      if (comma) g95_status(", ");
      g95_status("read="); g95_show_expr(i->read); comma = 1;
    }

    if (i->write) {
      if (comma) g95_status(", ");
      g95_status("write="); g95_show_expr(i->write); comma = 1;
    }

    if (i->readwrite) {
      if (comma) g95_status(", ");
      g95_status("readwrite="); g95_show_expr(i->readwrite); comma = 1;
    }

    if (i->delim) {
      if (comma) g95_status(", ");
      g95_status("delim="); g95_show_expr(i->delim); comma = 1;
    }

    if (i->pad) {
      if (comma) g95_status(", ");
      g95_status("pad="); g95_show_expr(i->pad); comma = 1;
    }

    if (i->err != NULL) {
      if (comma) g95_status(", ");
      g95_status("err=%d", i->err->value); comma = 1;
    }

    break;

  case EXEC_IOLENGTH:
    g95_status("iolength(expr=");
    g95_show_expr(c->expr);
    comma = 0;
    break;

  case EXEC_READ:
    g95_status("read(expr=");
    goto show_dt;

  case EXEC_WRITE:
    g95_status("write(expr=");

  show_dt:
    dt = c->ext.dt;
    comma = 0;

    if (dt->io_unit) {
      if (comma) g95_status(", ");
      g95_status("unit="); g95_show_expr(dt->io_unit); comma = 1;
    }

    if (dt->format_expr) {
      if (comma) g95_status(", ");
      g95_status("format_expr="); g95_show_expr(dt->format_expr); comma = 1;
    }

    if (dt->format_label != NULL) {
      if (comma) g95_status(", ");
      g95_status("format_label=%d", dt->format_label->value); comma = 1;
    }

    if (dt->namelist != NULL) {
      if (comma) g95_status(", ");
      g95_status("nml=%s", dt->namelist->name); comma = 1;
    }

    if (dt->iostat != NULL) {
      if (comma) g95_status(", ");
      g95_status("iostat="); g95_show_expr(dt->iostat); comma = 1;
    }

    if (dt->size != NULL) {
      if (comma) g95_status(", ");
      g95_status("size="); g95_show_expr(dt->size); comma = 1;
    }

    if (dt->rec != NULL) {
      if (comma) g95_status(", ");
      g95_status("rec="); g95_show_expr(dt->rec); comma = 1;
    }

    if (dt->advance) {
      if (comma) g95_status(", ");
      g95_status("advance="); g95_show_expr(dt->advance); comma = 1;
    }

    break;

  case EXEC_TRANSFER:
    g95_status("transfer(expr=");
    g95_show_expr(c->expr);
    comma = 1;
    break;

  case EXEC_DT_END:
    g95_status("dt_end(");
    comma = 0;

    dt = c->ext.dt;
    if (dt != NULL) {
      if (dt->err != NULL) {
	if (comma) g95_status(", ");
	g95_status("err=%d", dt->err->value); comma = 1;
      }

      if (dt->end != NULL) {
	if (comma) g95_status(", ");
	g95_status("end=%d", dt->end->value); comma = 1;
      }

      if (dt->eor != NULL) {
	if (comma) g95_status(", ");
	g95_status("eor=%d", dt->eor->value); comma = 1;
      }
    }

    break;

  case EXEC_ENTRY:
    g95_status("entry(name='%s'", c->sym->name);
    comma = 1;
    break;

  case EXEC_LABEL_ASSIGN:
    g95_status("assign_label(expr=");
    g95_show_expr(c->expr);
    g95_status(", label=%d", c->label->value);
    comma = 1;
    break;

  default:
    g95_internal_error("g95_show_code_node(): Bad statement code");
  }

  g95_status_char(')');
  g95_status_char('\n');
}



/* g95_show_code()-- Show a list of code structures. */

void g95_show_code(int level, g95_code *c) {

  for(;c ; c=c->next)
    g95_show_code_node(level, c);
}
